//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/From/GUIDomainSampleVisitor.cpp
//! @brief     Implements class GUIDomainSampleVisitor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/From/GUIDomainSampleVisitor.h"
#include "Base/Const/Units.h"
#include "GUI/Model/From/FromDomain.h"
#include "GUI/Model/Material/MaterialItem.h"
#include "GUI/Model/Material/MaterialModel.h"
#include "GUI/Model/Sample/FormFactorItems.h"
#include "GUI/Model/Sample/InterferenceItems.h"
#include "GUI/Model/Sample/LayerItem.h"
#include "GUI/Model/Sample/MesoCrystalItem.h"
#include "GUI/Model/Sample/MultiLayerItem.h"
#include "GUI/Model/Sample/ParticleCompositionItem.h"
#include "GUI/Model/Sample/ParticleCoreShellItem.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Model/Sample/ParticleLayoutItem.h"
#include "GUI/Model/Sample/SampleModel.h"
#include "GUI/Model/Trafo/RotationItems.h"
#include "GUI/Model/Trafo/TransformationItem.h"
#include "GUI/Model/Types/VectorItem.h"
#include "GUI/Util/Error.h"
#include "Param/Node/NodeUtils.h"
#include "Sample/Aggregate/ParticleLayout.h"
#include "Sample/HardParticle/HardParticles.h"
#include "Sample/Multilayer/Layer.h"
#include "Sample/Multilayer/MultiLayer.h"
#include "Sample/Multilayer/MultilayerUtils.h"
#include "Sample/Particle/Crystal.h"
#include "Sample/Particle/MesoCrystal.h"
#include "Sample/Particle/Particle.h"
#include "Sample/Particle/ParticleCoreShell.h"
#include "Sample/SoftParticle/SoftParticles.h"

#include <boost/polymorphic_cast.hpp>

using boost::polymorphic_cast;

namespace {

template <typename T> T* AddFormFactorItem(SessionItem* parent)
{
    if (parent->hasModelType<ParticleItem>())
        return polymorphic_cast<ParticleItem*>(parent)->setFormFactorType<T>();
    if (parent->hasModelType<MesoCrystalItem>())
        return polymorphic_cast<MesoCrystalItem*>(parent)->setOuterShapeType<T>();
    ASSERT(0);
}

} // namespace

//  ************************************************************************************************
//  class implementation
//  ************************************************************************************************

GUIDomainSampleVisitor::GUIDomainSampleVisitor() : m_sampleModel(nullptr), m_materialModel(nullptr)
{
}

GUIDomainSampleVisitor::~GUIDomainSampleVisitor() = default;

SessionItem* GUIDomainSampleVisitor::populateSampleModel(SampleModel* sampleModel,
                                                         MaterialModel* materialModel,
                                                         const MultiLayer& node,
                                                         const QString& node_name)
{
    m_sampleModel = sampleModel;
    m_materialModel = materialModel;
    m_levelToParentItem.clear();

    m_topSampleName = node_name;
    if (m_topSampleName.isEmpty())
        m_topSampleName = node.sampleName().c_str();

    for (const auto& [child, depth, parent] : NodeUtils::progenyPlus(&node)) {
        setDepth(depth + 1);
        child->accept(this); // will call back one of the visit functions defined below
    }
    SessionItem* result = m_levelToParentItem[1];

    // initialize all items with materials to find the materials
    for (auto multiLayerItem : sampleModel->multiLayerItems())
        for (auto itemWithMaterial : multiLayerItem->itemsWithMaterial())
            itemWithMaterial->setMaterialModel(materialModel);

    result->setItemName(m_topSampleName);
    return result;
}

//  ------------------------------------------------------------------------------------------------
//  visit functions
//  ------------------------------------------------------------------------------------------------

void GUIDomainSampleVisitor::visit(const ParticleLayout* node)
{
    SessionItem* parent = m_levelToParentItem[depth() - 1];
    auto* layout_item = m_sampleModel->insertItem<ParticleLayoutItem>(parent);
    layout_item->totalDensity().set(node->totalParticleSurfaceDensity());
    layout_item->weight().set(node->weight());
    m_levelToParentItem[depth()] = layout_item;
}

void GUIDomainSampleVisitor::visit(const Layer* node)
{
    SessionItem* parent = m_levelToParentItem[depth() - 1];
    ASSERT(parent);

    const auto* multilayer = dynamic_cast<const MultiLayer*>(m_itemToSample[parent]);
    ASSERT(multilayer);
    size_t i_layer = SampleUtils::Multilayer::IndexOfLayer(*multilayer, node);
    const LayerInterface* top_interface =
        i_layer == 0 ? nullptr : multilayer->layerInterface(i_layer - 1);

    auto* layer_item = m_sampleModel->insertItem<LayerItem>(parent);
    layer_item->setMaterial(createMaterialFromDomain(node->material()));

    GUI::Transform::FromDomain::setLayerItem(layer_item, node, top_interface);

    m_levelToParentItem[depth()] = layer_item;
}

void GUIDomainSampleVisitor::visit(const MultiLayer* node)
{
    auto* multilayer_item = m_sampleModel->insertItem<MultiLayerItem>();
    multilayer_item->setItemName(node->sampleName().c_str());
    multilayer_item->crossCorrLength().set(node->crossCorrLength());
    multilayer_item->setExternalField(node->externalField());
    m_levelToParentItem[depth()] = multilayer_item;
    m_itemToSample[multilayer_item] = node;
}

void GUIDomainSampleVisitor::visit(const Particle* node)
{
    auto* particle_item = InsertIParticle<ParticleItem>(node);
    particle_item->setMaterial(createMaterialFromDomain(node->material()));
}

void GUIDomainSampleVisitor::visit(const ParticleCoreShell* node)
{
    InsertIParticle<ParticleCoreShellItem>(node);
}

void GUIDomainSampleVisitor::visit(const ParticleComposition* node)
{
    InsertIParticle<ParticleCompositionItem>(node);
}

void GUIDomainSampleVisitor::visit(const MesoCrystal* node)
{
    InsertIParticle<MesoCrystalItem>(node);
}

void GUIDomainSampleVisitor::visit(const Crystal* node)
{
    SessionItem* item = m_levelToParentItem[depth() - 1];
    ASSERT(item);
    if (!item->hasModelType<MesoCrystalItem>()) {
        throw Error("GUI::Model::ObjectBuilder::visit(const Crystal*) "
                    "-> Error. Parent is not a MesoCrystal");
    }
    auto* mesocrystal_item = polymorphic_cast<MesoCrystalItem*>(item);
    auto lattice = node->transformedLattice();
    auto vector_a = lattice.getBasisVectorA();
    auto vector_b = lattice.getBasisVectorB();
    auto vector_c = lattice.getBasisVectorC();

    mesocrystal_item->setVectorA(vector_a);
    mesocrystal_item->setVectorB(vector_b);
    mesocrystal_item->setVectorC(vector_c);

    // Since there is no CrystalItem, set the parent map to the MesoCrystalItem
    m_levelToParentItem[depth()] = mesocrystal_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorAnisoPyramid* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<AnisoPyramidItem>(particle_item);
    ff_item->setLength(node->getLength());
    ff_item->setWidth(node->getWidth());
    ff_item->setHeight(node->getHeight());
    ff_item->setAlpha(Units::rad2deg(node->getAlpha()));
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorBarGauss* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<BarGaussItem>(particle_item);
    ff_item->setLength(node->getLength());
    ff_item->setWidth(node->getWidth());
    ff_item->setHeight(node->getHeight());
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorBarLorentz* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<BarLorentzItem>(particle_item);
    ff_item->setLength(node->getLength());
    ff_item->setWidth(node->getWidth());
    ff_item->setHeight(node->getHeight());
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorBox* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<BoxItem>(particle_item);
    ff_item->setLength(node->getLength());
    ff_item->setWidth(node->getWidth());
    ff_item->setHeight(node->getHeight());
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorCone* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<ConeItem>(particle_item);
    ff_item->setRadius(node->getRadius());
    ff_item->setHeight(node->getHeight());
    ff_item->setAlpha(Units::rad2deg(node->getAlpha()));
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorCone6* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<Cone6Item>(particle_item);
    ff_item->setBaseEdge(node->getBaseEdge());
    ff_item->setHeight(node->getHeight());
    ff_item->setAlpha(Units::rad2deg(node->getAlpha()));
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorCuboctahedron* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<CuboctahedronItem>(particle_item);
    ff_item->setLength(node->getLength());
    ff_item->setHeight(node->getHeight());
    ff_item->setHeightRatio(node->getHeightRatio());
    ff_item->setAlpha(Units::rad2deg(node->getAlpha()));
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorCylinder* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<CylinderItem>(particle_item);
    ff_item->setRadius(node->getRadius());
    ff_item->setHeight(node->getHeight());
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorDodecahedron* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<DodecahedronItem>(particle_item);
    ff_item->setEdge(node->getEdge());
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorEllipsoidalCylinder* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<EllipsoidalCylinderItem>(particle_item);
    ff_item->setRadiusX(node->getRadiusX());
    ff_item->setRadiusY(node->getRadiusY());
    ff_item->setHeight(node->getHeight());
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorFullSphere* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<FullSphereItem>(particle_item);
    ff_item->setRadius(node->getRadius());
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorFullSpheroid* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<FullSpheroidItem>(particle_item);
    ff_item->setRadius(node->getRadius());
    ff_item->setHeight(node->getHeight());
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorIcosahedron* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<IcosahedronItem>(particle_item);
    ff_item->setEdge(node->getEdge());
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorHemiEllipsoid* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<HemiEllipsoidItem>(particle_item);
    ff_item->setRadiusX(node->getRadiusX());
    ff_item->setRadiusY(node->getRadiusY());
    ff_item->setHeight(node->getHeight());
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorPrism3* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<Prism3Item>(particle_item);
    ff_item->setBaseEdge(node->getBaseEdge());
    ff_item->setHeight(node->getHeight());
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorPrism6* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<Prism6Item>(particle_item);
    ff_item->setBaseEdge(node->getBaseEdge());
    ff_item->setHeight(node->getHeight());
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorPyramid* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<PyramidItem>(particle_item);
    ff_item->setBaseEdge(node->getBaseEdge());
    ff_item->setHeight(node->getHeight());
    ff_item->setAlpha(Units::rad2deg(node->getAlpha()));
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorCosineRippleBox* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<CosineRippleBoxItem>(particle_item);
    ff_item->setLength(node->getLength());
    ff_item->setWidth(node->getWidth());
    ff_item->setHeight(node->getHeight());
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorCosineRippleGauss* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<CosineRippleGaussItem>(particle_item);
    ff_item->setLength(node->getLength());
    ff_item->setWidth(node->getWidth());
    ff_item->setHeight(node->getHeight());
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorCosineRippleLorentz* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<CosineRippleLorentzItem>(particle_item);
    ff_item->setLength(node->getLength());
    ff_item->setWidth(node->getWidth());
    ff_item->setHeight(node->getHeight());
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorSawtoothRippleBox* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<SawtoothRippleBoxItem>(particle_item);
    ff_item->setLength(node->getLength());
    ff_item->setWidth(node->getWidth());
    ff_item->setHeight(node->getHeight());
    ff_item->setAsymmetry(node->getAsymmetry());
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorSawtoothRippleGauss* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<SawtoothRippleGaussItem>(particle_item);
    ff_item->setLength(node->getLength());
    ff_item->setWidth(node->getWidth());
    ff_item->setHeight(node->getHeight());
    ff_item->setAsymmetry(node->getAsymmetry());
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorSawtoothRippleLorentz* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<SawtoothRippleLorentzItem>(particle_item);
    ff_item->setLength(node->getLength());
    ff_item->setWidth(node->getWidth());
    ff_item->setHeight(node->getHeight());
    ff_item->setAsymmetry(node->getAsymmetry());
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorTetrahedron* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<TetrahedronItem>(particle_item);
    ff_item->setBaseEdge(node->getBaseEdge());
    ff_item->setHeight(node->getHeight());
    ff_item->setAlpha(Units::rad2deg(node->getAlpha()));
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorTruncatedCube* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<TruncatedCubeItem>(particle_item);
    ff_item->setLength(node->getLength());
    ff_item->setRemovedLength(node->getRemovedLength());
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorTruncatedSphere* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<TruncatedSphereItem>(particle_item);
    ff_item->setRadius(node->getRadius());
    ff_item->setHeight(node->getHeight());
    ff_item->setRemovedTop(node->getRemovedTop());
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const FormFactorTruncatedSpheroid* node)
{
    SessionItem* particle_item = m_levelToParentItem[depth() - 1];
    auto* ff_item = AddFormFactorItem<TruncatedSpheroidItem>(particle_item);
    ff_item->setRadius(node->getRadius());
    ff_item->setHeight(node->getHeight());
    ff_item->setHeightFlattening(node->getHeightFlattening());
    ff_item->setRemovedTop(node->getRemovedTop());
    m_levelToParentItem[depth()] = particle_item;
}

void GUIDomainSampleVisitor::visit(const Interference1DLattice* node)
{
    ParticleLayoutItem* parent =
        dynamic_cast<ParticleLayoutItem*>(m_levelToParentItem[depth() - 1]);
    ASSERT(parent);
    auto* item = parent->createInterference<Interference1DLatticeItem>();
    GUI::Transform::FromDomain::set1DLatticeItem(item, *node);
    m_levelToParentItem[depth()] = item;
}

void GUIDomainSampleVisitor::visit(const Interference2DLattice* node)
{
    ParticleLayoutItem* parent =
        dynamic_cast<ParticleLayoutItem*>(m_levelToParentItem[depth() - 1]);
    ASSERT(parent);
    auto* item = parent->createInterference<Interference2DLatticeItem>();
    GUI::Transform::FromDomain::set2DLatticeItem(item, *node);
    m_levelToParentItem[depth()] = item;
}

void GUIDomainSampleVisitor::visit(const Interference2DParaCrystal* node)
{
    ParticleLayoutItem* parent =
        dynamic_cast<ParticleLayoutItem*>(m_levelToParentItem[depth() - 1]);
    ASSERT(parent);
    auto* item = parent->createInterference<Interference2DParaCrystalItem>();
    GUI::Transform::FromDomain::set2DParaCrystalItem(item, *node);
    m_levelToParentItem[depth()] = item;
}

void GUIDomainSampleVisitor::visit(const InterferenceFinite2DLattice* node)
{
    ParticleLayoutItem* parent =
        dynamic_cast<ParticleLayoutItem*>(m_levelToParentItem[depth() - 1]);
    ASSERT(parent);
    auto* item = parent->createInterference<InterferenceFinite2DLatticeItem>();
    GUI::Transform::FromDomain::setFinite2DLatticeItem(item, *node);
    m_levelToParentItem[depth()] = item;
}

void GUIDomainSampleVisitor::visit(const InterferenceHardDisk* node)
{
    ParticleLayoutItem* parent =
        dynamic_cast<ParticleLayoutItem*>(m_levelToParentItem[depth() - 1]);
    ASSERT(parent);
    auto* item = parent->createInterference<InterferenceHardDiskItem>();
    GUI::Transform::FromDomain::setHardDiskItem(item, *node);
    m_levelToParentItem[depth()] = item;
}

void GUIDomainSampleVisitor::visit(const InterferenceRadialParaCrystal* node)
{
    ParticleLayoutItem* parent =
        dynamic_cast<ParticleLayoutItem*>(m_levelToParentItem[depth() - 1]);
    ASSERT(parent);
    auto* item = parent->createInterference<InterferenceRadialParaCrystalItem>();
    GUI::Transform::FromDomain::setRadialParaCrystalItem(item, *node);
    m_levelToParentItem[depth()] = item;
}

void GUIDomainSampleVisitor::visit(const RotationX* node)
{
    auto* parent = dynamic_cast<ItemWithParticles*>(m_levelToParentItem[depth() - 1]);
    ASSERT(parent);

    TransformationItem* transformation_item = parent->createTransformationItem();
    auto* rot_item = transformation_item->setRotationType<XRotationItem>();
    rot_item->angle().set(Units::rad2deg(node->getAngle()));
    m_levelToParentItem[depth()] = transformation_item;
}

void GUIDomainSampleVisitor::visit(const RotationY* node)
{
    auto* parent = dynamic_cast<ItemWithParticles*>(m_levelToParentItem[depth() - 1]);
    ASSERT(parent);

    TransformationItem* transformation_item = parent->createTransformationItem();
    auto* rot_item = transformation_item->setRotationType<YRotationItem>();
    rot_item->angle().set(Units::rad2deg(node->getAngle()));
    m_levelToParentItem[depth()] = transformation_item;
}

void GUIDomainSampleVisitor::visit(const RotationZ* node)
{
    auto* parent = dynamic_cast<ItemWithParticles*>(m_levelToParentItem[depth() - 1]);
    ASSERT(parent);

    TransformationItem* transformation_item = parent->createTransformationItem();
    auto* rot_item = transformation_item->setRotationType<ZRotationItem>();
    rot_item->angle().set(Units::rad2deg(node->getAngle()));
    m_levelToParentItem[depth()] = transformation_item;
}

void GUIDomainSampleVisitor::visit(const RotationEuler* node)
{
    auto* parent = dynamic_cast<ItemWithParticles*>(m_levelToParentItem[depth() - 1]);
    ASSERT(parent);

    TransformationItem* transformation_item = parent->createTransformationItem();
    auto* rot_item = transformation_item->setRotationType<EulerRotationItem>();
    rot_item->alpha().set(Units::rad2deg(node->getAlpha()));
    rot_item->beta().set(Units::rad2deg(node->getBeta()));
    rot_item->gamma().set(Units::rad2deg(node->getGamma()));
    m_levelToParentItem[depth()] = transformation_item;
}

//  ------------------------------------------------------------------------------------------------
//  private functions for use in the above visit functions
//  ------------------------------------------------------------------------------------------------

MaterialItem* GUIDomainSampleVisitor::createMaterialFromDomain(const Material* material)
{
    QString materialName =
        m_topSampleName + QString("_") + QString(material->materialName().c_str());

    if (auto* materialItem = m_materialModel->materialFromName(materialName))
        return materialItem;

    complex_t material_data = material->materialData();
    MaterialItem* materialItem(nullptr);
    if (material->typeID() == MATERIAL_TYPES::RefractiveMaterial) {
        materialItem = m_materialModel->addRefractiveMaterial(materialName, material_data.real(),
                                                              material_data.imag());
    } else if (material->typeID() == MATERIAL_TYPES::MaterialBySLD) {
        materialItem = m_materialModel->addSLDMaterial(materialName, material_data.real(),
                                                       material_data.imag());
    } else
        ASSERT(0);

    materialItem->setMagnetization(material->magnetization());
    return materialItem;
}

// Function CreateIParticle, used only by GUIDomainSampleVisitor::InsertIParticle

// Default definition for any T, except if overwritten below
template <typename T>
T* GUIDomainSampleVisitor::CreateIParticle(SessionItem* parent, const IParticle*)
{
    return m_sampleModel->insertItem<T>(parent, -1);
}

// Special definition for T=ParticleItem
template <>
ParticleItem* GUIDomainSampleVisitor::CreateIParticle(SessionItem* parent,
                                                      const IParticle* particle)
{
    if (parent->hasModelType<ParticleCoreShellItem>()) {
        const auto* coreshell = dynamic_cast<const ParticleCoreShell*>(m_itemToSample[parent]);
        ASSERT(coreshell);
        auto* parentCoreShell = polymorphic_cast<ParticleCoreShellItem*>(parent);
        if (particle == coreshell->coreParticle())
            return parentCoreShell->createCore(m_materialModel);
        if (particle == coreshell->shellParticle())
            return parentCoreShell->createShell(m_materialModel);
        ASSERT(0);
    }
    return m_sampleModel->insertItem<ParticleItem>(parent, -1);
}

// Used only in this file
template <typename T> T* GUIDomainSampleVisitor::InsertIParticle(const IParticle* particle)
{
    auto* parent = m_levelToParentItem[depth() - 1];
    ASSERT(parent);

    T* particle_item = CreateIParticle<T>(parent, particle);

    ASSERT(particle_item);
    particle_item->setAbundance(particle->abundance());
    particle_item->setPosition(particle->position());

    m_levelToParentItem[depth()] = particle_item;
    m_itemToSample[particle_item] = particle;

    return particle_item;
}
