//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/ItemWithParticles.cpp
//! @brief     Implements class ItemWithParticles
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/ItemWithParticles.h"
#include "Base/Vector/Transform3D.h"
#include "GUI/Model/Session/SessionItemUtils.h"
#include "GUI/Model/Session/SessionModel.h"
#include "GUI/Model/Trafo/RotationItems.h"
#include "GUI/Model/Trafo/TransformationItem.h"
#include "GUI/Model/Types/VectorDescriptor.h"
#include "MesoCrystalItem.h"
#include "ParticleCompositionItem.h"
#include "ParticleCoreShellItem.h"
#include "Sample/Particle/IParticle.h"
#include "Sample/Scattering/Rotations.h"

namespace {

bool hasOwnAbundance(const SessionItem* item)
{
    static QStringList special_parent{ParticleCoreShellItem::M_TYPE,
                                      ParticleCompositionItem::M_TYPE, MesoCrystalItem::M_TYPE};

    return item ? special_parent.contains(item->modelType()) : false;
}

} // namespace

DoubleDescriptor ItemWithParticles::abundance() const
{
    return DoubleDescriptor(getItem(P_ABUNDANCE), Unit::unitless);
}

void ItemWithParticles::setAbundance(const double abundance)
{
    setItemValue(P_ABUNDANCE, abundance);
}

SessionItem* ItemWithParticles::abundanceItem() const
{
    return getItem(P_ABUNDANCE);
}

R3 ItemWithParticles::position() const
{
    return item<VectorItem>(P_POSITION)->getVector();
}

void ItemWithParticles::setPosition(const R3& position)
{
    item<VectorItem>(P_POSITION)->setVector(position);
}

VectorDescriptor ItemWithParticles::positionVector() const
{
    return VectorDescriptor(item<VectorItem>(P_POSITION), Unit::nanometer);
}

VectorItem* ItemWithParticles::positionItem() const
{
    return item<VectorItem>(P_POSITION);
}

TransformationItem* ItemWithParticles::createTransformationItem()
{
    return model()->insertItem<TransformationItem>(this, -1, T_TRANSFORMATION);
}

void ItemWithParticles::setTransformation(RotationItem* transformation)
{
    model()->moveItem(transformation, this, -1, T_TRANSFORMATION);
}

RotationItem* ItemWithParticles::rotationItem() const
{
    auto* transformationItem = dynamic_cast<TransformationItem*>(getItem(T_TRANSFORMATION));
    return transformationItem ? transformationItem->rotationItem() : nullptr;
}

bool ItemWithParticles::isTransformationTagName(const QString& name)
{
    return name == T_TRANSFORMATION;
}

Transform3D ItemWithParticles::rotation() const
{
    auto* const item = rotationItem();
    return item ? item->rotation() : Transform3D();
}

SelectionDescriptor<RotationItem*> ItemWithParticles::rotationMethod()
{
    SelectionDescriptor<RotationItem*> d;

    // we need a special filling for this selection descriptor (not just from a GroupItem), since
    // the rotation is stored in a TransformationItem instance, which can be present or not.

    static QVector<QPair<QString, QString>> map = {{"None", ""},
                                                   {"X axis Rotation", XRotationItem::M_TYPE},
                                                   {"Y axis Rotation", YRotationItem::M_TYPE},
                                                   {"Z axis Rotation", ZRotationItem::M_TYPE},
                                                   {"Euler Rotation", EulerRotationItem::M_TYPE}};

    d.label = "Rotation";

    for (auto [title, type] : map)
        d.options << title;

    d.currentItem = [=]() -> RotationItem* { return rotationItem(); };

    d.currentIndexSetter = [=](int current) {
        if (auto* item = getItem(T_TRANSFORMATION))
            model()->removeItem(item);
        if (current > 0)
            createTransformationItem()->setRotationType(map[current].second);
    };

    d.currentIndexGetter = [=]() {
        auto* item = rotationItem();
        if (item == nullptr)
            return 0;
        for (int i = 1; i < map.size(); i++)
            if (map[i].second == item->modelType())
                return i;

        return 0;
    };

    return d;
}

ItemWithParticles::ItemWithParticles(const QString& model_type, const QString& abundance_tooltip,
                                     const QString& position_tooltip)
    : SessionItem(model_type)
{
    addProperty(P_ABUNDANCE, 1.0)
        ->setLimits(RealLimits::limited(0.0, 1.0))
        .setDecimals(3)
        .setToolTip(abundance_tooltip);
    addProperty<VectorItem>(P_POSITION)->setToolTip(position_tooltip);

    registerTag(T_TRANSFORMATION, 0, 1, {TransformationItem::M_TYPE});

    mapper()->setOnParentChange([this](SessionItem* parent) {
        if (hasOwnAbundance(parent)) {
            setItemValue(P_ABUNDANCE, 1.0);
            getItem(P_ABUNDANCE)->setEnabled(false);
            if (isShellParticle()) {
                auto* pos = item<VectorItem>(P_POSITION);
                pos->setVector(R3());
                pos->setEnabled(false);
            }
        } else {
            getItem(P_ABUNDANCE)->setEnabled(true);
            if (isShellParticle())
                item<VectorItem>(P_POSITION)->setEnabled(true);
        }
    });
}

void ItemWithParticles::setDefaultTagTransformation()
{
    setDefaultTag(T_TRANSFORMATION);
}

bool ItemWithParticles::isShellParticle() const
{
    return false;
}

void ItemWithParticles::setTransformationInfo(IParticle* result) const
{
    result->setPosition(position());
    const Transform3D r = rotation();
    if (!r.isIdentity()) {
        std::unique_ptr<IRotation> rotation(IRotation::createRotation(r));
        result->setRotation(*rotation);
    }
}
