#include "Device/Detector/RectangularDetector.h"
#include "Base/Const/Units.h"
#include "Base/Util/Algorithms.h"
#include "Core/Simulation/GISASSimulation.h"
#include "Device/Beam/Beam.h"
#include "Tests/GTestWrapper/google_test.h"
#include <iostream>
#include <memory>

class RectangularDetectorTest : public ::testing::Test {
protected:
    double phi(R3 k) { return k.phi() / Units::deg; }
    double alpha(R3 k) { return 90.0 - k.theta() / Units::deg; }

    bool isEqual(const R3 lhs, const R3 rhs)
    {
        bool is_equal = BaseUtils::algo::almostEqual(lhs.x(), rhs.x())
                        && BaseUtils::algo::almostEqual(lhs.y(), rhs.y())
                        && BaseUtils::algo::almostEqual(lhs.z(), rhs.z());
        if (!is_equal)
            std::cout << "lhs:" << lhs << " rhs:" << rhs << " diff:" << (lhs - rhs) << std::endl;
        return is_equal;
    }
};

TEST_F(RectangularDetectorTest, InitialState)
{
    RectangularDetector det(50u, 10.0, 60u, 20.0);
    EXPECT_EQ(50u, det.getNbinsX());
    EXPECT_EQ(10.0, det.getWidth());
    EXPECT_EQ(60u, det.getNbinsY());
    EXPECT_EQ(20.0, det.getHeight());

    EXPECT_EQ(0.0, det.getU0());
    EXPECT_EQ(0.0, det.getV0());
    EXPECT_EQ(0.0, det.getDistance());

    EXPECT_TRUE(R3() == det.getNormalVector());
    EXPECT_TRUE(R3(0.0, -1.0, 0.0) == det.getDirectionVector());

    EXPECT_EQ(RectangularDetector::GENERIC, det.getDetectorArrangment());
}

TEST_F(RectangularDetectorTest, Clone)
{
    RectangularDetector det(50u, 10.0, 60u, 20.0);
    R3 normal(10.0, 20.0, 30.0);
    R3 direction(1.0, 2.0, 3.0);
    double u0(88.0), v0(99.0);
    det.setPosition(normal, u0, v0, direction);

    std::unique_ptr<RectangularDetector> clone(det.clone());
    EXPECT_EQ(u0, clone->getU0());
    EXPECT_EQ(v0, clone->getV0());
    EXPECT_TRUE(normal == clone->getNormalVector());
    EXPECT_TRUE(direction == clone->getDirectionVector());
    EXPECT_EQ(RectangularDetector::GENERIC, clone->getDetectorArrangment());
}

TEST_F(RectangularDetectorTest, PerpToSample)
{
    size_t nbinsx(5u), nbinsy(4u);
    double width(50.0), height(40.0);
    double distance(100.0), u0(20.0), v0(10.0);
    //    double dx = width / nbinsx;
    //    double dy = height / nbinsy;

    RectangularDetector det(nbinsx, width, nbinsy, height);

    // detector perpendicular to sample
    det.setPerpendicularToSampleX(distance, u0, v0);
    EXPECT_EQ(distance, det.getDistance());
    EXPECT_EQ(u0, det.getU0());
    EXPECT_EQ(v0, det.getV0());
    EXPECT_TRUE(R3() == det.getNormalVector());
    EXPECT_TRUE(R3(0.0, -1.0, 0.0) == det.getDirectionVector());
    EXPECT_EQ(RectangularDetector::PERPENDICULAR_TO_SAMPLE, det.getDetectorArrangment());

    // initializing with the simulation
    GISASSimulation simulation;
    simulation.setBeamParameters(1.0, 10.0 * Units::deg, 0.0);
    det.setDetectorNormal(simulation.beam().direction().zReflected());
    EXPECT_TRUE(R3(distance, 0, 0) == det.getNormalVector());
    EXPECT_TRUE(R3(0.0, -1.0, 0.0) == det.getDirectionVector());
}

TEST_F(RectangularDetectorTest, PerpToDirectBeam)
{
    size_t nbinsx(5u), nbinsy(4u);
    double width(50.0), height(40.0);
    double distance(100.0), u0(20.0), v0(10.0);
    //    double dx = width / nbinsx;
    //    double dy = height / nbinsy;
    double alpha_i(10.0 * Units::deg);

    RectangularDetector det(nbinsx, width, nbinsy, height);

    // detector perpendicular to direct beam
    det.setPerpendicularToDirectBeam(distance, u0, v0);
    EXPECT_EQ(distance, det.getDistance());
    EXPECT_EQ(u0, det.getU0());
    EXPECT_EQ(v0, det.getV0());
    EXPECT_TRUE(R3() == det.getNormalVector());
    EXPECT_TRUE(R3(0.0, -1.0, 0.0) == det.getDirectionVector());
    EXPECT_EQ(RectangularDetector::PERPENDICULAR_TO_DIRECT_BEAM, det.getDetectorArrangment());

    // initializing with the simulation
    GISASSimulation simulation;
    simulation.setBeamParameters(1.0, alpha_i, 0.0);
    det.setDetectorNormal(simulation.beam().direction().zReflected());
    R3 normal(distance * cos(alpha_i), 0.0, -1.0 * distance * sin(alpha_i));
    EXPECT_TRUE(isEqual(normal, det.getNormalVector()));
    EXPECT_TRUE(R3(0.0, -1.0, 0.0) == det.getDirectionVector());
}

TEST_F(RectangularDetectorTest, PerpToReflectedBeam)
{
    size_t nbinsx(5u), nbinsy(4u);
    double width(50.0), height(40.0);
    double distance(100.0), u0(20.0), v0(10.0);
    //    double dx = width / nbinsx;
    //    double dy = height / nbinsy;
    double alpha_i(10.0 * Units::deg);

    RectangularDetector det(nbinsx, width, nbinsy, height);

    // detector perpendicular to reflected beam
    det.setPerpendicularToReflectedBeam(distance, u0, v0);
    EXPECT_EQ(distance, det.getDistance());
    EXPECT_EQ(u0, det.getU0());
    EXPECT_EQ(v0, det.getV0());
    EXPECT_TRUE(R3() == det.getNormalVector());
    EXPECT_TRUE(R3(0.0, -1.0, 0.0) == det.getDirectionVector());
    EXPECT_EQ(RectangularDetector::PERPENDICULAR_TO_REFLECTED_BEAM, det.getDetectorArrangment());

    // initializing with the simulation
    GISASSimulation simulation;
    simulation.setBeamParameters(1.0, alpha_i, 0.0);
    det.setDetectorNormal(simulation.beam().direction().zReflected());
    R3 normal(distance * cos(alpha_i), 0.0, 1.0 * distance * sin(alpha_i));
    EXPECT_TRUE(isEqual(normal, det.getNormalVector()));
    EXPECT_TRUE(R3(0.0, -1.0, 0.0) == det.getDirectionVector());
}

// detector perpendicular to reflected beam, when direct beam position is known
TEST_F(RectangularDetectorTest, PerpToReflectedBeamDpos)
{
    size_t nbinsx(5u), nbinsy(4u);
    double width(50.0), height(40.0);
    double distance(100.0), u0(20.0), v0(10.0);
    //    double dx = width / nbinsx;
    //    double dy = height / nbinsy;
    double alpha_i(10.0 * Units::deg);

    RectangularDetector det(nbinsx, width, nbinsy, height);

    // detector perpendicular to reflected beam
    det.setPerpendicularToReflectedBeam(distance);
    EXPECT_EQ(distance, det.getDistance());
    EXPECT_EQ(0.0, det.getU0());
    EXPECT_EQ(0.0, det.getV0());
    EXPECT_EQ(0.0, det.getDirectBeamU0());
    EXPECT_EQ(0.0, det.getDirectBeamV0());

    double dbeam_u0 = u0;
    double dbeam_v0 = -distance * std::tan(alpha_i * 2.0) + v0;
    det.setDirectBeamPosition(dbeam_u0, dbeam_v0);

    EXPECT_EQ(dbeam_u0, det.getDirectBeamU0());
    EXPECT_EQ(dbeam_v0, det.getDirectBeamV0());
    EXPECT_TRUE(R3() == det.getNormalVector());
    EXPECT_TRUE(R3(0.0, -1.0, 0.0) == det.getDirectionVector());
    EXPECT_EQ(RectangularDetector::PERPENDICULAR_TO_REFLECTED_BEAM_DPOS,
              det.getDetectorArrangment());

    // initializing with the simulation
    GISASSimulation simulation;
    simulation.setBeamParameters(1.0, alpha_i, 0.0);
    det.setDetectorNormal(simulation.beam().direction().zReflected());

    R3 normal(distance * cos(alpha_i), 0.0, 1.0 * distance * sin(alpha_i));
    EXPECT_TRUE(isEqual(normal, det.getNormalVector()));
    EXPECT_TRUE(R3(0.0, -1.0, 0.0) == det.getDirectionVector());
    EXPECT_EQ(u0, det.getU0());
    EXPECT_EQ(v0, det.getV0());
}

// Test retrieval of analyzer properties
TEST_F(RectangularDetectorTest, AnalyzerProperties)
{
    RectangularDetector detector(50u, 10.0, 60u, 20.0);

    R3 direction;
    double efficiency = 0.0;
    double total_transmission = 1.0;
    R3 unit_direction;

    // if direction is the zero vector, an exception is thrown
    EXPECT_THROW(detector.setAnalyzer(direction, efficiency, total_transmission),
                 std::runtime_error);

    // zero efficiency
    direction = R3(1.0, 0.0, 0.0);
    unit_direction = direction.unit();
    detector.setAnalyzer(direction, efficiency, total_transmission);
    const PolFilter& detect_properties = detector.analyzer();

    EXPECT_NEAR(detect_properties.polEfficiency(), efficiency, 1e-8);
    EXPECT_NEAR(detect_properties.totalTransmission(), total_transmission, 1e-8);
    // direction vector returned is zero vector because efficiency is zero
    EXPECT_NEAR(detect_properties.polDirection().x(), 0.0, 1e-8);
    EXPECT_NEAR(detect_properties.polDirection().y(), 0.0, 1e-8);
    EXPECT_NEAR(detect_properties.polDirection().z(), 0.0, 1e-8);

    // intermediate efficiency
    direction = R3(1.0, 0.0, 0.0);
    efficiency = 0.5;
    total_transmission = 0.6;
    unit_direction = direction.unit();
    detector.setAnalyzer(direction, efficiency, total_transmission);
    const PolFilter& detect_properties2 = detector.analyzer();

    EXPECT_NEAR(detect_properties2.polEfficiency(), efficiency, 1e-8);
    EXPECT_NEAR(detect_properties2.totalTransmission(), total_transmission, 1e-8);
    EXPECT_NEAR(detect_properties2.polDirection().x(), unit_direction.x(), 1e-8);
    EXPECT_NEAR(detect_properties2.polDirection().y(), unit_direction.y(), 1e-8);
    EXPECT_NEAR(detect_properties2.polDirection().z(), unit_direction.z(), 1e-8);

    // maximum efficiency
    direction = R3(1.0, 0.0, 0.0);
    efficiency = 1.0;
    total_transmission = 0.5;
    unit_direction = direction.unit();
    detector.setAnalyzer(direction, efficiency, total_transmission);
    const PolFilter& detect_properties3 = detector.analyzer();

    EXPECT_NEAR(detect_properties3.polEfficiency(), efficiency, 1e-8);
    EXPECT_NEAR(detect_properties3.totalTransmission(), total_transmission, 1e-8);
    EXPECT_NEAR(detect_properties3.polDirection().x(), unit_direction.x(), 1e-8);
    EXPECT_NEAR(detect_properties3.polDirection().y(), unit_direction.y(), 1e-8);
    EXPECT_NEAR(detect_properties3.polDirection().z(), unit_direction.z(), 1e-8);

    // non-axis direction
    direction = R3(1.0, 2.0, 3.0);
    efficiency = 1.0;
    total_transmission = 0.5;
    unit_direction = direction.unit();
    detector.setAnalyzer(direction, efficiency, total_transmission);
    const PolFilter& detect_properties4 = detector.analyzer();

    EXPECT_NEAR(detect_properties4.polEfficiency(), efficiency, 1e-8);
    EXPECT_NEAR(detect_properties4.totalTransmission(), total_transmission, 1e-8);
    EXPECT_NEAR(detect_properties4.polDirection().x(), unit_direction.x(), 1e-8);
    EXPECT_NEAR(detect_properties4.polDirection().y(), unit_direction.y(), 1e-8);
    EXPECT_NEAR(detect_properties4.polDirection().z(), unit_direction.z(), 1e-8);

    // maximum efficiency and negative efficiency
    direction = R3(0.0, -1.0, -1.0);
    efficiency = -1.0;
    total_transmission = 0.5;
    unit_direction = direction.unit();
    detector.setAnalyzer(direction, efficiency, total_transmission);
    const PolFilter& detect_properties5 = detector.analyzer();

    EXPECT_NEAR(detect_properties5.polEfficiency(), efficiency, 1e-8);
    EXPECT_NEAR(detect_properties5.totalTransmission(), total_transmission, 1e-8);
    EXPECT_NEAR(detect_properties5.polDirection().x(), unit_direction.x(), 1e-8);
    EXPECT_NEAR(detect_properties5.polDirection().y(), unit_direction.y(), 1e-8);
    EXPECT_NEAR(detect_properties5.polDirection().z(), unit_direction.z(), 1e-8);
}
